"""Timestamp-related utility functions."""

from .timestamp import (
    timestamp_to_local_time,
    timestamp_to_local_time_str,
    get_timestamp,
    timestamp_to_datetime,
    tz_aware_dt_from_timestamp_and_tz,
    timestamp_to_dateint,
)
try:
    del timestamp
except NameError: # pragma: no cover
    pass



"""Timestamp-related utility functions."""

from datetime import datetime
import calendar

import pytz
from delorean import Delorean

from ..datetime import datetime_to_dateint


def timestamp_to_local_time(timestamp, timezone_name):
    """Convert epoch timestamp to a localized Delorean datetime object.

    Arguments
    ---------
    timestamp : int
        The timestamp to convert.
    timezone_name : datetime.timezone
        The timezone of the desired local time.

    Returns
    -------
    delorean.Delorean
        A localized Delorean datetime object.
    """
    # first convert timestamp to UTC
    utc_time = datetime.utcfromtimestamp(float(timestamp))
    delo = Delorean(utc_time, timezone='UTC')
    # shift d according to input timezone
    localized_d = delo.shift(timezone_name)
    return localized_d


def timestamp_to_local_time_str(
        timestamp, timezone_name, fmt="yyyy-MM-dd HH:mm:ss"):
    """Convert epoch timestamp to a localized datetime string.

    Arguments
    ---------
    timestamp : int
        The timestamp to convert.
    timezone_name : datetime.timezone
        The timezone of the desired local time.
    fmt : str
        The format of the output string.

    Returns
    -------
    str
        The localized datetime string.
    """
    localized_d = timestamp_to_local_time(timestamp, timezone_name)
    localized_datetime_str = localized_d.format_datetime(fmt)
    return localized_datetime_str


def get_timestamp(timezone_name, year, month, day, hour=0, minute=0):
    """Epoch timestamp from timezone, year, month, day, hour and minute."""
    tz = pytz.timezone(timezone_name)
    tz_datetime = tz.localize(datetime(year, month, day, hour, minute))
    timestamp = calendar.timegm(tz_datetime.utctimetuple())
    return timestamp


def timestamp_to_datetime(timestamp):
    """Converts a UTC timestamp to a UTC-aligned datetime object.

    Arguments
    ---------
    timestamp : int
        A UTC timestamp.

    Returns
    -------
    datetime.datetime
        A UTC-aligned datetime object corresponding to the given timestamp.
    """
    return datetime.utcfromtimestamp(timestamp)


def tz_aware_dt_from_timestamp_and_tz(timestamp, timezone_name):
    """Creates a timezone-aware datetime object from given timestamp and
    timezone."""
    return datetime.fromtimestamp(timestamp, timezone_name)


def timestamp_to_dateint(timestamp):
    """Converts a UTC timestamp to a dateint of the corresponding day.

    Arguments
    ---------
    timestamp : int
        A UTC timestamp.

    Returns
    -------
    int
        An integer object decipting the calendaric day - e.g. 20161225 -
        corresponding to the given timestamp.
    """
    return datetime_to_dateint(timestamp_to_datetime(timestamp))


_LEAP_YEAR_SINCE_EPOCH = [1972, 1976, 1980, 1984, 1988, 1992, 1996, 2000, 2004,
                          2008, 2012, 2016]

_AVG_SEC_IN_YEAR = 365 * 24 * 60 * 60 + 5 * 60 * 60 + 48 * 60 + 45

# from ..constants import (SECONDS_IN_COMMON_YEAR, SECONDS_IN_LEAP_YEAR)

def _efficient_timestamp_to_dateint():
    #todo: use above constants
    pass



""" Usefull time-related constants."""

SECONDS_IN_MINUTE = 60
MINUTES_IN_HOUR = 60
HOURS_IN_DAY = 24
SECONDS_IN_HOUR = MINUTES_IN_HOUR * SECONDS_IN_MINUTE
MINUTES_IN_DAY = HOURS_IN_DAY * MINUTES_IN_HOUR
SECONDS_IN_DAY = MINUTES_IN_DAY * SECONDS_IN_MINUTE
SECONDS_IN_COMMON_YEAR = 31536000
SECONDS_IN_LEAP_YEAR = 31622400
DAYS_IN_A_WEEK = 7

WEEKDAYS = [
    'Monday',
    'Tuesday',
    'Wednesday',
    'Thursday',
    'Friday',
    'Saturday',
    'Sunday'
]

WESTERN_WORKDAYS = [
    'Monday',
    'Tuesday',
    'Wednesday',
    'Thursday',
    'Friday'
]

JEWISH_WORKDAYS = [
    'Sunday',
    'Monday',
    'Tuesday',
    'Wednesday',
    'Thursday'
]



"""Dateint-related utility functions."""

from .dateint import (
    decompose_dateint,
    dateint_to_date,
    tz_aware_dateint_to_timestamp,
    dateint_to_timestamp,
    dateint_to_utc_timestamp,
    dateint_to_datetime,
    dateint_to_weekday,
    dateint_to_weekday_name,
    shift_dateint,
    dateint_range,
    today_int,
    dateint_week_by_dateint,
    dateint_difference,
)
try:
    del dateint
except NameError: # pragma: no cover
    pass



"""Datetime-related utility functions."""


class TimeInterval:
    """A class that represents a time interval, based on a start and end point.
    """
    def __init__(self, start, end):
        """Create a new TimeInterval object from a start and end point.

        Parameters
        ----------
        start, end : datetime.datetime
        """
        assert start <= end
        self.start = start
        self.end = end

    @classmethod
    def from_timedelta(cls, datetime_obj, duration):
        """Create a new TimeInterval object from a start point and a duration.

        If duration is positive, datetime_obj is the start of the interval;
        if duration is negative, datetime_obj is the end of the interval.

        Parameters
        ----------
        datetime_obj : datetime.datetime
        duration : datetime.timedelta

        Returns
        -------
        neutils.time.TimeInterval
        """
        if duration.total_seconds() > 0:
            return TimeInterval(datetime_obj, datetime_obj + duration)
        else:
            return TimeInterval(datetime_obj + duration, datetime_obj)

    def __repr__(self):
        return "{}({!r}, {!r})".format(
            type(self).__name__, self.start, self.end)

    def __str__(self):
        return "{} -> {}".format(self.start, self.end)

    def __contains__(self, datetime_obj):
        """Check if a certain datetime objects is in the interval, requiring
        that it is between the start and end points (inclusive).

        Parameters
        ----------
        datetime_obj: datetime.datetime

        Returns
        -------
        bool
        """
        return self.start <= datetime_obj <= self.end



"""Utility pure-Python 3 decorators."""

import utilitime.dateint
import utilitime.datetime
import utilitime.time
import utilitime.timestamp
import utilitime.weekday

from .time_interval import TimeInterval
try:
    del utilitime
    del time_interval
except NameError: # pragma: no cover
    pass

# pylint: disable=C0413
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



"""Datetime-related utility functions."""

from datetime import datetime, timedelta, date
import math

from ..timestamp import get_timestamp
from ..datetime import (
    utc_time,
    datetime_to_dateint,
)
from ..constants import (
    WEEKDAYS,
)


def decompose_dateint(dateint):
    """Decomposes the given dateint into its year, month and day components.

    Arguments
    ---------
    dateint : int
        An integer object decipting a specific calendaric day; e.g. 20161225.

    Returns
    -------
    year : int
        The year component of the given dateint.
    month : int
        The month component of the given dateint.
    day : int
        The day component of the given dateint.
    """
    year = int(dateint / 10000)
    leftover = dateint - year * 10000
    month = int(leftover / 100)
    day = leftover - month * 100
    return year, month, day


def dateint_to_date(dateint):
    """Converts the given integer to a datetime.date object.

    Arguments
    ---------
    dateint : int
        An integer object decipting a specific calendaric day; e.g. 20161225.

    Returns
    -------
    datetime.date
        The corresponding date object.

    Example
    -------
    >>> dateint_to_date(20170223)
    datetime.date(2017, 2, 23)
    """
    return date(*decompose_dateint(dateint))


def tz_aware_dateint_to_timestamp(dateint, timezone_name):
    """Returns the epoch timestamp for the given timezone and dateint.

    Arguments
    ---------
    dateint : int
        An integer object decipting a specific calendaric day; e.g. 20161225.
    timezone_name : str
        The name of the timezone.

    Returns
    -------
    int
        The timestamp corresponding to the start of the given day (so at 0
        hours, 0 minutes, etc...) at the given timezone.
    """
    return get_timestamp(timezone_name, *decompose_dateint(dateint))


def dateint_to_timestamp(dateint):
    """Converts the given dateint to a timestamp, using the local timezone.

    Arguments
    ---------
    dateint : int
        An integer object decipting a specific calendaric day; e.g. 20161225.

    Returns
    -------
    int
        The timestamp corresponding to the start of the given day (so at 0
        hours, 0 minutes, etc...) at the local timezone.
    """
    return int(dateint_to_datetime(dateint).timestamp())


def dateint_to_utc_timestamp(dateint):
    """Converts the given dateint to the corresponding UTC timestamp.

    Arguments
    ---------
    dateint : int
        An integer object decipting a specific calendaric day; e.g. 20161225.

    Returns
    -------
    int
        The UTC timestamp corresponding to the start of the given day (so at 0
        hours, 0 minutes, etc...).
    """
    return tz_aware_dateint_to_timestamp(dateint, 'UTC')


def dateint_to_datetime(dateint):
    """Converts the given dateint to a datetime object, in local timezone.

    Arguments
    ---------
    dateint : int
        An integer object decipting a specific calendaric day; e.g. 20161225.

    Returns
    -------
    datetime.datetime
        A timezone-unaware datetime object representing the start of the given
        day (so at 0 hours, 0 minutes, etc...) in the local timezone.
    """
    if len(str(dateint)) != 8:
        raise ValueError(
            'Dateints must have exactly 8 digits; the first four representing '
            'the year, the next two the months, and the last two the days.')
    year, month, day = decompose_dateint(dateint)
    return datetime(year=year, month=month, day=day)


def dateint_to_weekday(dateint, first_day='Monday'):
    """Returns the weekday of the given dateint.

    Arguments
    ---------
    dateint : int
        An integer object decipting a specific calendaric day; e.g. 20161225.
    first_day : str, default 'Monday'
        The first day of the week.

    Returns
    -------
    int
        The weekday of the given dateint, when first day of the week = 0,
        last day of the week = 6.

    Example
    -------
    >>> dateint_to_weekday(20170213)
    0
    >>> dateint_to_weekday(20170212)
    6
    >>> dateint_to_weekday(20170214)
    1
    >>> dateint_to_weekday(20170212, 'Sunday)
    0
    >>> dateint_to_weekday(20170214, 'Sunday')
    2
    """
    weekday_ix = dateint_to_datetime(dateint).weekday()
    return (weekday_ix - WEEKDAYS.index(first_day)) % 7


def dateint_to_weekday_name(dateint):
    """Returns the weekday of the given dateint.

    Arguments
    ---------
    dateint : int
        An integer object decipting a specific calendaric day; e.g. 20161225.

    Returns
    -------
    str
        The weekday name of the given dateint.

    Example
    -------
    >>> dateint_to_weekday_name(20170213)
    'Monday'
    >>> dateint_to_weekday_name(20170212)
    'Sunday'
    >>> dateint_to_weekday_name(20170214)
    'Tuesday'
    """
    return dateint_to_datetime(dateint).strftime("%A")


def shift_dateint(dateint, day_shift):
    """Shifts the given dateint by the given amount of days.

    Arguments
    ---------
    dateint : int
        An integer object decipting a specific calendaric day; e.g. 20161225.
    days : int
        The number of days to shift the given dateint by. A negative number
        shifts the dateint backwards.

    Returns
    -------
    int
       A dateint corresponding to the given date shifted by the given amount
       of days.

    Example
    -------
    >>> shift_dateint(20170228, 1)
    20170301
    >>> shift_dateint(20170301, -1)
    20170228
    >>> shift_dateint(20170220, 5)
    20170225
    """
    dtime = dateint_to_datetime(dateint)
    delta = timedelta(days=abs(day_shift))
    if day_shift > 0:
        dtime = dtime + delta
    else:
        dtime = dtime - delta
    return datetime_to_dateint(dtime)


def dateint_range(first_dateint, last_dateint):
    """Returns all dateints in the given dateint range.

    Arguments
    ---------
    first_dateint : int
        An integer object decipting a specific calendaric day; e.g. 20161225.
    last_dateint : int
        An integer object decipting a specific calendaric day; e.g. 20170108.

    Returns
    -------
    iterable
        An iterable of ints representing all days in the given dateint range.

    Example
    -------
    >>> dateint_range(20170228, 20170301)
    [20170228, 20170301]
    >>> dateint_range(20170225, 20170301)
    [20170225, 20170226, 20170227, 20170228, 20170301]
    """
    first_datetime = dateint_to_datetime(first_dateint)
    last_datetime = dateint_to_datetime(last_dateint)
    delta = last_datetime - first_datetime
    delta_in_hours = math.ceil(delta.total_seconds() / 3600)
    delta_in_days = math.ceil(delta_in_hours / 24) + 1
    dateint_set = set()
    for delta_i in range(0, delta_in_days * 24, 24):
        datetime_i = first_datetime + timedelta(hours=delta_i)
        dateint_i = datetime_to_dateint(datetime_i)
        if dateint_i <= last_dateint:
            dateint_set.add(dateint_i)
    return sorted(dateint_set)


def today_int():
    """Returns the dateint for today."""
    return datetime_to_dateint(utc_time())


def dateint_week_by_dateint(dateint, first_day='Monday'):
    """Return a dateint range of the week the given dateint belongs to.

    Arguments
    ---------
    dateint : int
        An integer object decipting a specific calendaric day; e.g. 20161225.
    first_day : str, default 'Monday'
        The first day of the week.

    Returns
    -------
    iterable
        An iterable of dateint representing all days of the week the given
        dateint belongs to.
    """
    weekday_ix = dateint_to_weekday(dateint, first_day)
    first_day_dateint = shift_dateint(dateint, -weekday_ix)
    last_day_dateint = shift_dateint(first_day_dateint, 6)
    return dateint_range(first_day_dateint, last_day_dateint)


def dateint_difference(dateint1, dateint2):
    """Return the difference between two dateints in days.

    Arguments
    ---------
    dateint1 : int
        An integer object decipting a specific calendaric day; e.g. 20161225.
    dateint2 : int
        An integer object decipting a specific calendaric day; e.g. 20161225.

    Returns
    -------
    int
        The difference between the two given dateints in days.
    """
    dt1 = dateint_to_datetime(dateint1)
    dt2 = dateint_to_datetime(dateint2)
    delta = dt1 - dt2
    return abs(delta.days)



"""Weekday-related utility functions."""

from decore import lazy_property

from ..constants import (
    WEEKDAYS,
)


# === weekday-related functions ===

def next_weekday(weekday):
    """Returns the name of the weekday after the given weekday name."""
    ix = WEEKDAYS.index(weekday)
    if ix == len(WEEKDAYS)-1:
        return WEEKDAYS[0]
    return WEEKDAYS[ix+1]


def prev_weekday(weekday):
    """Returns the name of the weekday before the given weekday name."""
    ix = WEEKDAYS.index(weekday)
    if ix == 0:
        return WEEKDAYS[len(WEEKDAYS)-1]
    return WEEKDAYS[ix-1]


@lazy_property
def _lower_weekdays():
    return [day.lower() for day in WEEKDAYS]


@lazy_property
def _double_weekdays():
    return WEEKDAYS + WEEKDAYS


def workdays(first_day=None):
    """Returns a list of workday names.

    Arguments
    ---------
    first_day : str, default None
        The first day of the five-day work week. If not given, 'Monday' is
        used.

    Returns
    -------
    list
        A list of workday names.
    """
    if first_day is None:
        first_day = 'Monday'
    ix = _lower_weekdays().index(first_day.lower())
    return _double_weekdays()[ix:ix+5]


def weekdays(first_day=None):
    """Returns a list of weekday names.

    Arguments
    ---------
    first_day : str, default None
        The first day of the week. If not given, 'Monday' is used.

    Returns
    -------
    list
        A list of weekday names.
    """
    if first_day is None:
        first_day = 'Monday'
    ix = _lower_weekdays().index(first_day.lower())
    return _double_weekdays()[ix:ix+7]



"""dateime.time-related utility functions."""

from datetime import time

from ..constants import (
    SECONDS_IN_DAY,
)


def decompose_seconds_in_day(seconds):
    """Decomposes seconds in day into hour, minute and second components.

    Arguments
    ---------
    seconds : int
        A time of day by the number of seconds passed since midnight.

    Returns
    -------
    hour : int
        The hour component of the given time of day.
    minut : int
        The minute component of the given time of day.
    second : int
        The second component of the given time of day.
    """
    if seconds > SECONDS_IN_DAY:
        seconds = seconds - SECONDS_IN_DAY
    if seconds < 0:
        raise ValueError("seconds param must be non-negative!")
    hour = int(seconds / 3600)
    leftover = seconds - hour * 3600
    minute = int(leftover / 60)
    second = leftover - minute * 60
    return hour, minute, second


def seconds_in_day_to_time(seconds):
    """Decomposes atime of day into hour, minute and seconds components.

    Arguments
    ---------
    seconds : int
        A time of day by the number of seconds passed since midnight.

    Returns
    -------
    datetime.time
        The corresponding time of day as a datetime.time object.

    Example
    -------
    >>> seconds_in_day_to_time(23430)
    datetime.time(6, 30, 30)
    """
    try:
        return time(*decompose_seconds_in_day(seconds))
    except ValueError:
        print("Seconds = {}".format(seconds))
        print("H = {}, M={}, S={}".format(*decompose_seconds_in_day(seconds)))
        raise


def minutes_in_day_to_time(minutes):
    """Decomposes atime of day into hour, minute and seconds components.

    Arguments
    ---------
    minutes : int
        A time of day by the number of minutes passed since midnight.

    Returns
    -------
    datetime.time
        The corresponding time of day as a datetime.time object.

    Example
    -------
    >>> minutes_in_day_to_time(390)
    datetime.time(6, 30, 00)
    """
    return seconds_in_day_to_time(minutes*60)



"""Datetime-related utility functions."""

from .datetime import (
    epoch_datetime,
    utc_time,
    utc_offset_by_timezone,
    localize_datetime,
    datetime_to_dateint,
    local_datetime_to_timestamp,
    utc_datetime_to_timestamp,
)
try:
    del datetime
except NameError: # pragma: no cover
    pass



"""Datetime-related utility functions."""

from datetime import datetime, timezone

import pytz
from decore import lazy_property

from ..constants import (
    SECONDS_IN_HOUR,
)


# === datetime-related functions ===

@lazy_property
def epoch_datetime():
    """Returns the epoch as a datetime.datetime object."""
    return datetime.utcfromtimestamp(0)


def utc_time():
    """Returns the current server time as a datetime object.

    This assumes all servers run in UTC time.
    """
    return datetime.utcnow()


def utc_offset_by_timezone(timezone_name):
    """Returns the UTC offset of the given timezone in hours.

    Arguments
    ---------
    timezone_name: str
        A string with a name of a timezone.

    Returns
    -------
    int
        The UTC offset of the given timezone, in hours.
    """
    return int(pytz.timezone(timezone_name).utcoffset(
        utc_time()).total_seconds()/SECONDS_IN_HOUR)


def localize_datetime(datetime_obj, timezone_name):
    """Localizes the given UTC-aligned datetime by the given timezone.

    Arguments
    ---------
    datetime_obj : datetime.datetime
        A datetime object decipting a specific point in time, aligned by UTC.
    timezone_name: str
        A string with a name of a timezone.

    Returns
    -------
    datetime.datetime
        An datetime object aligned by the given timezone.
    """
    return datetime_obj.replace(tzinfo=pytz.utc).astimezone(
        pytz.timezone(timezone_name))


def datetime_to_dateint(datetime_obj):
    """Converts the given datetime object to the corresponding dateint.

    Arguments
    ---------
    datetime_obj : datetime.datetime
        A datetime object decipting a specific point in time.

    Returns
    -------
    int
        An integer represeting the day, month and year of the given point in
        time. For example, 3:32 AM on December 3rd 2015 will be converted to
        the integer 20151203.
    """
    return datetime_obj.year * 10000 + datetime_obj.month * 100 \
        + datetime_obj.day


def local_datetime_to_timestamp(datetime_obj):
    """Converts the given localized naive datetime object to a UTC timestamp.

    Arguments
    ---------
    datetime_obj : datetime.datetime
        A naive (not timezone-aware) datetime object decipting a specific
        point in time in the local machine timezone.

    Returns
    -------
    int
        The UTC timestamp corresponding to the given datetime object.
    """
    return int(datetime_obj.timestamp())


def utc_datetime_to_timestamp(datetime_obj):
    """Converts the given naive UTC-aligned datetime object to a UTC timestamp.

    Arguments
    ---------
    datetime_obj : datetime.datetime
        A naive (not timezone-aware) datetime object decipting a specific
        point in time in UTC time.

    Returns
    -------
    int
        The UTC timestamp corresponding to the given datetime object.
    """
    return int(datetime_obj.replace(tzinfo=timezone.utc).timestamp())



"""datetime.time-related utility functions."""

from .time import (
    decompose_seconds_in_day,
    seconds_in_day_to_time,
    minutes_in_day_to_time,
)
try:
    del time
except NameError: # pragma: no cover
    pass
